<?php

namespace Drupal\entity_mesh\Form;

use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Menu\MenuTreeStorageInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Configure Entity Mesh settings for this site.
 */
class SettingsForm extends ConfigFormBase {

  /**
   * Config settings.
   *
   * @var string
   */
  const SETTINGS = 'entity_mesh.settings';

  /**
   * Supported Entities.
   *
   * @var array<string>
   */
  const SUPPORTED_ENTITIES = [
    'node',
    'taxonomy_term',
    'user',
  ];

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $database;

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Entity\EntityTypeBundleInfoInterface
   */
  protected $entityTypeBundleInfo;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The menu tree storage.
   *
   * @var \Drupal\Core\Menu\MenuTreeStorageInterface
   */
  protected $menuTreeStorage;

  /**
   * Constructor.
   *
   * @param \Drupal\Core\Database\Connection $database
   *   The database connection.
   * @param \Drupal\Core\Entity\EntityTypeBundleInfoInterface $entity_type_bundle_info
   *   The entity type bundle info service.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager service.
   * @param \Drupal\Core\Menu\MenuTreeStorageInterface $menu_tree_storage
   *   The menu tree storage.
   */
  final public function __construct(Connection $database, EntityTypeBundleInfoInterface $entity_type_bundle_info, EntityTypeManagerInterface $entity_type_manager, MenuTreeStorageInterface $menu_tree_storage) {
    $this->database = $database;
    $this->entityTypeBundleInfo = $entity_type_bundle_info;
    $this->entityTypeManager = $entity_type_manager;
    $this->menuTreeStorage = $menu_tree_storage;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('database'),
      $container->get('entity_type.bundle.info'),
      $container->get('entity_type.manager'),
      $container->get(MenuTreeStorageInterface::class)
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'entity_mesh_settings';
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() {
    return ['entity_mesh.settings'];
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $config = $this->config(self::SETTINGS);

    $form['global'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Global'),
    ];

    $form['global']['external'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('External URLs'),
      '#default_value' => $config->get('external'),
    ];

    $form['global']['internal'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Internal URLs'),
      '#default_value' => $config->get('internal'),
    ];
    $form['global']['process_mesh_using_queue'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Recalculated dependencies using queue'),
      '#description' => $this->t('When an operation is performed on an entity,
      such as a node, its dependencies are recalculated. There is the possibility
      that this recalculation is executed within the same workflow as the operation
      or added to a queue to be executed later.'),
      '#default_value' => $config->get('process_mesh_using_queue'),
    ];

    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $this->config(self::SETTINGS)
      ->set('external', $form_state->getValue('external'))
      ->set('internal', $form_state->getValue('internal'))
      ->set('process_mesh_using_queue', $form_state->getValue('process_mesh_using_queue'))
      ->save();

    parent::submitForm($form, $form_state);
  }

}
